import jax
import jax.numpy as jnp
import jax.random as jr
from jax.scipy.stats import norm, dirichlet
from jaxtyping import Array, Float, Int
from jax.nn import softmax
import tqdm
from typing import Tuple
from functools import partial
from dynamax.hidden_markov_model import (
    hmm_filter,
    hmm_smoother,
    hmm_posterior_mode,
    hmm_posterior_sample,
    parallel_hmm_posterior_sample,
)

from .util import (
    simulate_markov_chain,
    lower_dim,
    raise_dim,
    count_transitions,
)

na = jnp.newaxis

"""
data = {
    "syllables": (n_sequences, n_timesteps, n_syllables),
    "mask": (n_sequences, n_timesteps),
}

states: (n_sequences, n_timesteps)

params = {
    "emissions": (n_states, n_syllables, n_syllables),
    "trans_probs": (n_states, n_states),
}

hypparams = {
    "n_states": (,),
    "emission_beta": (,),
    "trans_beta": (,),
    "trans_kappa": (,),
    "n_syllables"
}
"""


def obs_log_likelihoods(
    data: dict,
    params: dict,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Compute log likelihoods of observations for each hidden state."""

    n_sequences = data["syllables"].shape[0]
    n_states = params["trans_probs"].shape[0]

    log_syllable_trans_probs = jnp.log(params["emissions"])

    log_likelihoods = jax.vmap(
        lambda T: T[data["syllables"][:, :-1], data["syllables"][:, 1:]]
    )(log_syllable_trans_probs)

    log_likelihoods = jnp.concatenate(
        [jnp.zeros((n_states, n_sequences, 1)), log_likelihoods], axis=2
    )
    return log_likelihoods.transpose((1, 2, 0)) * data["mask"][:, :, na]


def log_params_prob(
    params: dict,
    hypparams: dict,
) -> Float:
    """Compute the log probability of the parameters based on their priors."""

    n_states, n_syllables = params["emissions"].shape[:2]

    # prior on emission base parameters
    emissions_log_prob = jax.vmap(dirichlet.logpdf, in_axes=(0, None))(
        params["emissions"].reshape(-1, n_syllables),
        jnp.ones(n_syllables) * hypparams["emissions_beta"],
    ).sum()

    # prior on transition parameters
    trans_probs_log_prob = jax.vmap(dirichlet.logpdf)(
        params["trans_probs"],
        jnp.eye(n_states) * hypparams["trans_kappa"] + hypparams["trans_beta"],
    ).sum()

    return emissions_log_prob + trans_probs_log_prob


def log_joint_prob(
    data: dict,
    params: dict,
    hypparams: dict,
) -> Float:
    """Compute the log joint probability of the data and parameters."""
    return marginal_loglik(data, params) + log_params_prob(params, hypparams)


@partial(jax.jit, static_argnums=(3,))
def resample_states(
    seed: Float[Array, "2"],
    data: dict,
    params: dict,
    parallel: bool = False,
) -> Tuple[Int[Array, "n_sequences n_timesteps"], Float]:
    """Resample hidden states from their posterior distribution.

    Args:
        seed: random seed
        data: data dictionary
        params: parameters dictionary
        parallel: whether to use parallel message passing

    Returns:
        states: resampled hidden states
        marginal_loglik: marginal log likelihood of the data
    """
    n_states = params["trans_probs"].shape[0]
    seeds = jr.split(seed, data["syllables"].shape[0])

    sample_fn = parallel_hmm_posterior_sample if parallel else hmm_posterior_sample
    marginal_logliks, states = jax.vmap(sample_fn, in_axes=(0, None, None, 0))(
        seeds,
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    )
    return states, marginal_logliks.sum()


def fit_gibbs(
    data: dict,
    hypparams: dict,
    init_params: dict,
    init_states: Int[Array, "n_sequences n_timesteps"] = None,
    seed: Float[Array, "2"] = jr.PRNGKey(0),
    num_iters: Int = 100,
    parallel: bool = False,
) -> Tuple[dict, Float[Array, "num_iters"]]:
    """Fit a model using Gibbs sampling.

    Args:
        data: data dictionary
        hypparams: hyperparameters dictionary
        init_params: initial parameters directionary
        init_states: initial hidden states (optional)
        seed: random seed
        num_iters: number of iterations
        parallel: whether to use parallel message passing

    Returns:
        params: fitted parameters dictionary
        log_joints: log joint probability of the data and parameters recorded at each iteration
    """
    if init_states is None:
        states, _ = resample_states(seed, data, init_params, parallel)

    log_joints = []
    params = init_params
    for _ in tqdm.trange(num_iters):
        seed, subseed = jr.split(seed)
        params = resample_params(subseed, data, params, states, hypparams)
        states, marginal_loglik = resample_states(seed, data, params, parallel)
        log_joints.append(marginal_loglik + log_params_prob(params, hypparams))
    return params, states, jnp.array(log_joints)


def initialize_params(
    data: dict,
    hypparams: dict,
    states: Int[Array, "n_sequences n_timesteps"] = None,
    seed: Float[Array, "2"] = jr.PRNGKey(0),
) -> dict:
    """Initialize parameters by sampling from their prior distribution or using
    provided states.

    Args:
        data: data dictionary
        hypparams: hyperparameters dictionary
        states: states used for initializing the parameters (optional)
        seed: random seed
    """
    if states is not None:
        params = resample_params(seed, data, states, hypparams)
    else:
        params = random_params(seed, hypparams)
    return params


def marginal_loglik(
    data: dict,
    params: dict,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Estimate marginal log likelihood of the data"""
    n_states = params["trans_probs"].shape[0]
    mll = jax.vmap(hmm_filter, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    ).marginal_loglik.sum()
    return mll


def smoothed_states(
    data: dict,
    params: dict,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Estimate marginals of hidden states using forward-backward algorithm."""
    n_states = params["trans_probs"].shape[0]
    return jax.vmap(hmm_smoother, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    ).smoothed_probs


def filtered_states(
    data: dict,
    params: dict,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Estimate marginals of hidden states using forward-backward algorithm."""
    n_states = params["trans_probs"].shape[0]
    return jax.vmap(hmm_filter, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    ).filtered_probs


def predicted_states(
    data: dict,
    params: dict,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Predict hidden states using Viterbi algorithm."""
    n_states = params["trans_probs"].shape[0]
    return jax.vmap(hmm_posterior_mode, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    )


def random_params(
    seed: Float[Array, "2"],
    hypparams: dict,
) -> dict:
    """Generate random model parameters.

    emissions ~ Dirichlet(emissions_beta) (for each state and syllable)
    trans_probs ~ Dirichlet(trans_beta + trans_kappa * I)

    Args:
        seed: random seed
        hypparams: hyperparameters dictionary

    Returns:
        params: parameters dictionary
    """
    n_syllables = hypparams["n_syllables"]
    n_states = hypparams["n_states"]
    seeds = jr.split(seed, 2)

    emissions = jr.dirichlet(
        seeds[0],
        jnp.ones((n_states, n_syllables, n_syllables)) * hypparams["emissions_beta"],
    )

    trans_probs = jax.vmap(jr.dirichlet)(
        jr.split(seeds[2], n_states),
        jnp.eye(n_states) * hypparams["trans_kappa"] + hypparams["trans_beta"],
    )

    return {
        "emissions": emissions,
        "trans_probs": trans_probs,
    }


def simulate(
    seed: Float[Array, "2"],
    params: dict,
    n_timesteps: Int,
    n_sequences: Int,
) -> Tuple[
    Int[Array, "n_sequences n_timesteps"], Int[Array, "n_sequences n_timesteps"]
]:
    """Simulate data from the model.

    Args:
        seed: random seed
        params: parameters dictionary
        n_timesteps: number of timesteps to simulate
        n_sequences: number of sessions to simulate

    Returns:
        states: simulated states
        syllables: simulated syllables
    """
    seeds = jr.split(seed, 3)

    states = jax.vmap(simulate_markov_chain, in_axes=(0, None, None))(
        jr.split(seeds[0], n_sequences),
        params["trans_probs"],
        n_timesteps,
    )

    syllable_trans_probs = params["emissions"][states]

    syllables = jax.vmap(simulate_markov_chain, in_axes=(0, 0, None))(
        jr.split(seeds[1], n_sequences), syllable_trans_probs, n_timesteps
    )
    return states, syllables


def resample_params(
    seed: Float[Array, "2"],
    data: dict,
    params: dict,
    states: Int[Array, "n_sequences n_timesteps"],
    hypparams: dict,
) -> dict:
    """Resample parameters from their posterior distribution.

    Args:
        seed: random seed
        data: data dictionary
        params: parameters dictionary
        states: hidden states
        hypparams: hyperparameters dictionary

    Returns:
        params: parameters dictionary
    """
    seeds = jr.split(seed, 2)

    emissions = resample_emission_params(
        seeds[1],
        data["syllables"],
        data["mask"],
        states,
        hypparams["n_states"],
        hypparams["n_syllables"],
        hypparams["emissions_beta"],
    )
    trans_probs = resample_trans_probs(
        seeds[2],
        data["mask"],
        states,
        hypparams["n_states"],
        hypparams["trans_beta"],
        hypparams["trans_kappa"],
    )
    params = {
        "emissions": emissions,
        "trans_probs": trans_probs,
    }
    return params

    emissions = resample_emission_params(
        seeds[1],
        data["syllables"],
        data["mask"],
        states,
        hypparams["n_states"],
        hypparams["n_syllables"],
        hypparams["emissions_beta"],
    )


@partial(jax.jit, static_argnums=(4, 5))
def resample_emission_params(
    seed: Float[Array, "2"],
    syllables: Int[Array, "n_sequences n_timesteps"],
    mask: Int[Array, "n_sequences n_timesteps"],
    states: Int[Array, "n_sequences n_timesteps"],
    n_states: int,
    n_syllables: int,
    emissions_beta: Float,
) -> Float[Array, "n_states n_syllables n_syllables"]:
    """Resample emission parameters from their posterior distribution.

    Args:
        seed: random seed
        syllables: syllable observations
        mask: mask of valid observations
        states: hidden states
        n_states: number of hidden states
        n_syllables: number of syllables
        emission_base_sigma: emission base standard deviation
        emission_biases_sigma: emission biases standard deviation

    Returns:
        emissions: syllable transition probabilities for each state
    """
    sufficient_stats = (
        jnp.zeros((n_states, n_syllables, n_syllables))
        .at[states[:, 1:], syllables[:, :-1], syllables[:, 1:]]
        .add(mask[:, 1:])
    )
    emissions = jax.vmap(jr.dirichlet)(
        jr.split(seed, n_states * n_syllables),
        sufficient_stats.reshape(-1, n_syllables) + emissions_beta,
    ).reshape(n_states, n_syllables, n_syllables)

    return emissions


@partial(jax.jit, static_argnums=(3,))
def resample_trans_probs(
    seed: Float[Array, "2"],
    mask: Int[Array, "n_sequences n_timesteps"],
    states: Int[Array, "n_sequences n_timesteps"],
    n_states: int,
    beta: Float,
    kappa: Float,
) -> Float[Array, "n_states n_states"]:
    """Resample transition probabilities from their posterior distribution.

    Args:
        seed: random seed
        mask: mask of valid observations
        states: hidden states
        n_states: number of hidden states
        beta: Dirichlet concentration parameter
        kappa: Dirichlet concentration parameter

    Returns:
        trans_probs: posterior transition probabilities
    """
    trans_counts = jax.vmap(count_transitions, in_axes=(0, 0, None))(states, mask, n_states).sum(0)
    trans_probs = jax.vmap(jr.dirichlet)(
        jr.split(seed, n_states), trans_counts + beta + jnp.eye(n_states) * kappa
    )
    return trans_probs



import os
import cv2
import imageio
import numpy as np
import tqdm
from vidio.read import OpenCVReader
from typing import Dict, Tuple, List, Union
from jaxtyping import Array, Float, Int
import plotly.graph_objects as go
import matplotlib.pyplot as plt
from .util import sample_instances, compare_states, get_frequencies

def crop_image(
    image: Array,
    centroid: Tuple[int, int],
    crop_size: Union[int, Tuple[int, int]],
) -> Array:
    """Crop an image around a centroid.

    Args:
        image: Image to crop, shape (height, width, channels).
        centroid: (x, y) coordinates of the centroid.
        crop_size: Size of crop around centroid (int or (width, height)).

    Returns:
        cropped_image: Cropped (and zero-padded if needed) image.
    """
    if isinstance(crop_size, tuple):
        w, h = crop_size
    else:
        w, h = crop_size, crop_size
    x, y = int(centroid[0]), int(centroid[1])

    x_min = max(0, x - w // 2)
    y_min = max(0, y - h // 2)
    x_max = min(image.shape[1], x + w // 2)
    y_max = min(image.shape[0], y + h // 2)

    cropped = image[y_min:y_max, x_min:x_max]
    padded = np.zeros((h, w, *image.shape[2:]), dtype=image.dtype)
    pad_x = max(w // 2 - x, 0)
    pad_y = max(h // 2 - y, 0)
    padded[pad_y:pad_y+cropped.shape[0], pad_x:pad_x+cropped.shape[1]] = cropped
    return padded


def write_video_clip(
    frames: Array,
    path: str,
    fps: int = 30,
    quality: int = 7,
) -> None:
    """Write a video clip to a file using imageio.

    Args:
        frames: Video frames of shape (n_frames, height, width, channels).
        path: Output file path.
        fps: Frames per second.
        quality: Video encoding quality.
    """
    with imageio.get_writer(path, pixelformat="yuv420p", fps=fps, quality=quality) as writer:
        for frame in frames:
            writer.append_data(frame)


def grid_movie(
    instances: List[Tuple[str, int, int]],
    rows: int,
    cols: int,
    videos: Dict,
    centroids: Dict[str, Float[Array, "n_timesteps 2"]],
    window_size: int,
    max_duration: int = 1200,
    dimming_factor: float = 0.5,
) -> Int[Array, "n_frames window_size*rows window_size*cols 3"]:
    """Generate a grid movie from video instances.

    Args:
        instances: List of (key, start, end) tuples.
        rows: Number of rows in grid.
        cols: Number of columns in grid.
        videos: Dictionary of video readers keyed by name.
        centroids: Dictionary of centroids for each frame.
        window_size: Size of cropped window around centroid.
        max_duration: Maximum duration of the grid movie.
        dimming_factor: Multiplicative factor for dimming after state instance has ended.

    Returns:
        frames: Combined movie frames as a single array.
    """
    tiles = []
    instance_durations = [end - start for _, start, end in instances]
    max_duration = min(max_duration, max(instance_durations))
    for key, start, end in instances:
        tile = []
        video_length = len(videos[key])
        for t in range(start, start + max_duration):
            if t >= video_length:
                frame = np.zeros((window_size, window_size, 3), np.uint8)
            else:
                cen = centroids[key][t]
                frame = crop_image(videos[key][t], cen, window_size)
                if t >= end:
                    frame = np.uint8(frame * dimming_factor)
                frame[:,0] = 100
                frame[:,-1] = 100
                frame[0,:] = 100
                frame[-1,:] = 100
            tile.append(frame)
        tiles.append(np.stack(tile))

    for _ in range(rows * cols - len(tiles)):
        tiles.append(np.zeros_like(tiles[0]))

    tiles = np.stack(tiles).reshape(rows, cols, max_duration, window_size, window_size, 3)
    frames = np.concatenate(np.concatenate(tiles, axis=2), axis=2)
    return frames


def generate_grid_movies(
    states_dict: Dict[str, Int[Array, "n_timesteps"]],
    video_paths: Dict[str, str],
    centroids: Dict[str, Float[Array, "n_timesteps 2"]],
    output_dir: str = "grid_movies",
    rows: int = 3,
    cols: int = 4,
    max_duration: int = 1200,
    dimming_factor: float = 0.25,
    quality: int = 7,
    window_size: int = 256,
) -> Dict[int, List[Tuple[str, int, int]]]:
    """Generate and save grid movies for high-level states.

    Args:
        states_dict: Dictionary of state sequences.
        video_paths: Dictionary of paths to video files.
        centroids: Dictionary of centroids for each frame.
        output_dir: Directory to save output movies.
        rows: Number of rows in each grid.
        cols: Number of columns in each grid.
        max_duration: Maximum duration of each movie.
        dimming_factor: Factor to dim frames after state instance has ended.
        quality: Video encoding quality.
        window_size: Size of window around centroid.

    Returns:
        instances: Sampled instances used for grid movies.
    """
    os.makedirs(output_dir, exist_ok=True)
    videos = {k: OpenCVReader(path) for k, path in video_paths.items()}
    fps = videos[list(video_paths.keys())[0]].fps

    instances = sample_instances(states_dict, rows * cols)
    for state_ix, inst_list in tqdm.tqdm(instances.items(), desc="Generating grid movies", ncols=72):
        frames = grid_movie(
            inst_list,
            rows,
            cols,
            videos,
            centroids,
            window_size,
            max_duration=max_duration,
            dimming_factor=dimming_factor,
        )
        path = os.path.join(output_dir, f"state{state_ix}.mp4")
        write_video_clip(frames, path, fps=fps, quality=quality)

    return instances



def plot_sankey(
    states_dict1: Dict[str, Int[Array, "n_timesteps"]],
    states_dict2: Dict[str, Int[Array, "n_timesteps"]],
) -> go.Figure:
    """Create a Sankey diagram representing the relationship between two sets of states.

    Args:
        states_dict1: First dictionary of state sequences.
        states_dict2: Second dictionary of state sequences.

    Returns:
        fig: Plotly figure object containing the Sankey diagram.
    """
    import plotly.io as pio
    pio.renderers.default = "notebook"

    confusion_matrix = compare_states(states_dict1, states_dict2)[0]
    frequencies1 = get_frequencies(states_dict1)
    frequencies2 = get_frequencies(states_dict2)
    n1, n2 = len(frequencies1), len(frequencies2)
    
    cmap = plt.get_cmap("tab20")
    node_colors = [f"rgba({int(r*255)},{int(g*255)},{int(b*255)},0.8)" 
                   for r, g, b, _ in [cmap(i % 20) for i in range(max(n1, n2))]]
    node_colors_combined = node_colors[:n1] + node_colors[:n2]
    node_labels = [f"state {i}" for i in range(n1)] + [f"state {i}" for i in range(n2)]

    sources, targets, values, link_colors = [], [], [], []
    for i in range(n1):
        for j in range(n2):
            val = confusion_matrix[i, j] * frequencies1[i]
            if val > 0:
                sources.append(i)
                targets.append(n1 + j)  # offset right side
                values.append(val)
                link_colors.append(node_colors[i])

    # Build Sankey
    fig = go.Figure(go.Sankey(
        arrangement="snap",
        node=dict(
            pad=15,
            thickness=20,
            line=dict(color="black", width=0.5),
            label=node_labels,
            color=node_colors_combined
        ),
        link=dict(
            source=sources,
            target=targets,
            value=values,
            color=link_colors
        )
    ))
    return fig


import jax.numpy as jnp
import jax.random as jr
import jax
import optax
import tqdm
import h5py
import numpy as np
from functools import partial
from tensorflow_probability.substrates.jax import distributions as tfd
from jaxtyping import Array, Float, Int, PyTree, Bool
from typing import Tuple, Union, Callable, Dict, Optional, List
from scipy.optimize import linear_sum_assignment
from dynamax.utils.optimize import run_gradient_descent
from jax.scipy.linalg import cho_factor, cho_solve
from sklearn.metrics import adjusted_rand_score

na = jnp.newaxis


@jax.vmap
def logits_to_probs(
    logits: Float[Array, "n_categories-1"]
) -> Float[Array, "n_categories"]:
    """Convert logits to probabilities."""
    logits = jnp.concatenate([logits, jnp.zeros(1)])
    return jax.nn.softmax(logits)


@jax.vmap
def probs_to_logits(
    probs: Float[Array, "n_categories"],
    pseudo_count: Float = 1e-8,
) -> Float[Array, "n_categories-1"]:
    """Convert probabilities to logits."""
    log_probs = jnp.log(probs + pseudo_count)
    return log_probs[:-1] - log_probs[-1]


def normal_inverse_gamma_posterior(
    seed: Float[Array, "2"],
    mean: Float,
    sigmasq: Float,
    n: Int,
    lambda_: Float,
    alpha: Float,
    beta: Float,
) -> Tuple[Float, Float]:
    """
    Sample posterior mean and variance given normal-inverse gamma prior.

    Args:
        seed: random seed
        mean: sample mean
        sigmasq: sample variance
        n: number of data points
        lambda_: strength of prior
        alpha: inverse gamma shape parameter
        beta: inverse gamma rate parameter

    Returns:
        mu: posterior mean
        sigma: posterior variance
    """
    seeds = jr.split(seed, 2)
    mean = jnp.nan_to_num(mean)
    sigmasq = jnp.nan_to_num(sigmasq)
    lambda_n = lambda_ + n
    alpha_n = alpha + n / 2
    beta_n = beta + 0.5 * n * sigmasq + 0.5 * n * lambda_ * (mean**2) / lambda_n
    sigma = sample_inv_gamma(seeds[0], alpha_n, beta_n)
    mu = jr.normal(seeds[1]) * jnp.sqrt(sigmasq / lambda_n) + mean
    return mu, sigma


def center_embedding(n: int) -> Float[Array, "n n-1"]:
    """Generate an orthonormal matrix that embeds R^(n-1) into the space of 0-sum vectors in R^n."""
    # using numpy.linalg.svd because jax version crashes on windows
    X = jnp.tril(jnp.ones((n, n)), k=-1)[1:]
    X = jnp.eye(n)[1:] - X / X.sum(1)[:, na]
    X = X / jnp.sqrt((X**2).sum(1))[:, na]
    return X.T


def lower_dim(arr, axis=0):
    """Lower dimension in specified axis by projecting onto the space of 0-sum vectors."""
    arr = jnp.moveaxis(arr, axis, 0)
    k, *shape = arr.shape
    arr = arr.reshape(k, -1)
    arr = center_embedding(k).T @ arr
    arr = arr.reshape(k - 1, *shape)
    arr = jnp.moveaxis(arr, 0, axis)
    return arr


def raise_dim(arr, axis=0):
    """Raise dimension in specified axis by embedding into the space of 0-sum vectors."""
    arr = jnp.moveaxis(arr, axis, 0)
    k, *shape = arr.shape
    arr = arr.reshape(k, -1)
    arr = center_embedding(k + 1) @ arr
    arr = arr.reshape(k + 1, *shape)
    arr = jnp.moveaxis(arr, 0, axis)
    return arr


def sample_multinomial(
    seed: Float[Array, "2"],
    n: Int,
    p: Float[jnp.ndarray, "n_categories"],
) -> Int[Array, "n_categories"]:
    return tfd.Multinomial(n, probs=p).sample(seed=seed)


def sample_gamma(
    seed: Float[Array, "2"],
    a: Float,
    b: Float,
) -> Float:
    return jr.gamma(seed, a) / b


def sample_inv_gamma(
    seed: Float[Array, "2"],
    a: Float,
    b: Float,
) -> Float:
    return 1.0 / sample_gamma(seed, a, b)


@partial(jax.jit, static_argnames=["n_timesteps"])
def simulate_markov_chain(
    seed: Float[Array, "2"],
    trans_probs: Union[
        Float[Array, "n_states n_states"], Float[Array, "n_timesteps n_states n_states"]
    ],
    n_timesteps: Int,
    init_probs: Optional[Float[Array, "n_states"]] = None,
) -> Int[Array, "n_timesteps"]:
    """Simulate a state sequence from in Markov chain.

    Args:
        seed: random seed
        trans_probs: transition probabilities between states
        n_timesteps: number of timesteps to simulate
        init_probs: initial state probabilities. If None, uniform distribution is used.

    Returns:
        states: simulated state sequence
    """
    seeds = jr.split(seed, n_timesteps + 1)
    n_states = trans_probs.shape[0]
    log_trans_probs = jnp.log(trans_probs)
    if init_probs is None:
        log_init_probs = jnp.zeros(n_states)
    else:
        log_init_probs = jnp.log(init_probs)
    init_state = jr.categorical(seeds[0], log_init_probs)

    if trans_probs.ndim == 2:
        def step(state, seed):
            next_state = jr.categorical(seed, log_trans_probs[state])
            return next_state, next_state
        _, states = jax.lax.scan(step, init_state, seeds[1:])

    else:
        def step(state, args):
            seed, logT = args
            next_state = jr.categorical(seed, logT[state])
            return next_state, next_state
        _, states = jax.lax.scan(step, init_state, (seeds[1:], log_trans_probs))
        
    return states


def count_transitions(
    states: Int[Array, "n_timesteps"],
    mask: Int[Array, "n_timesteps"],
    n_states: int,
) -> Float[Array, "n_states n_states"]:
    """Count transitions between states.

    Args:
        states: discrete state sequence
        mask: mask of valid observations
        n_states: number of discrete states

    Returns:
        trans_counts: transition counts
    """
    trans_counts = (
        jnp.zeros((n_states, n_states))
        .at[states[:-1], states[1:]]
        .add(mask[:-1])
    )
    return trans_counts


def compare_states(
    states1: Union[Int[Array, "n_timesteps"], Dict[str, Int[Array, "n_timesteps"]]],
    states2: Union[Int[Array, "n_timesteps"], Dict[str, Int[Array, "n_timesteps"]]],
    n_states: Int = None,
) -> Tuple[Int[Array, "n_states n_states"], Int[Array, "n_states"], Float]:
    """Compare high-level state sequences.

    Args:
        states1: first set of state sequences (can be an array or dictionary of sequences)
        states2: second set of state sequences (can be an array or dictionary of sequences)
        n_states: number of discrete states. If None, inferred from data.

    Returns:
        confusion_matrix: confusion matrix
        optimal_permutation: optimal permutation of first set of states to match second set
        accuracy: proportion of timepoints with matching labels (after optimal permutation)
    """
    if isinstance(states1, dict):
        states1 = _concatenate_stateseqs(states1)
    if isinstance(states2, dict):
        states2 = _concatenate_stateseqs(states2)
    if n_states is None:
        n_states = max(states1.max(), states2.max()) + 1

    confusion = jnp.zeros((n_states, n_states)).at[states1, states2].add(1)
    optimal_perm = linear_sum_assignment(-confusion.T)[1]
    accuracy = confusion[optimal_perm, jnp.arange(n_states)].sum() / states2.size
    confusion = confusion / confusion.sum(axis=1, keepdims=True)
    return confusion, optimal_perm, accuracy


def sample_hmc(
    seed: Float[Array, "2"],
    log_prob_fn: Callable,
    init_params: PyTree,
    num_leapfrog_steps: Int = 3,
    step_size: Float = 0.001,
    num_results: Int = 1,
    num_burnin_steps: Int = 100,
) -> Tuple[PyTree, PyTree]:
    """Sample using Hamiltonian Monte Carlo."""
    hmc_kernel = tfp.mcmc.HamiltonianMonteCarlo(
        target_log_prob_fn=log_prob_fn,
        step_size=step_size,
        num_leapfrog_steps=num_leapfrog_steps,
    )
    params, _, kernel_state = tfp.mcmc.sample_chain(
        num_results=num_results,
        num_burnin_steps=num_burnin_steps,
        current_state=init_params,
        kernel=hmc_kernel,
        seed=seed,
        trace_fn=None,
        return_final_kernel_results=True,
    )
    return params, kernel_state


def sample_laplace(
    seed: Float[Array, "2"],
    log_prob_fn: Callable,
    init_params: PyTree,
    gradient_descent_iters: Int = 200,
    gradient_descent_lr: Float = 0.01,
) -> Tuple[PyTree, Float[Array, "gradient_descent_iters"]]:
    """Sample using Laplace approximation. Uses gradient descent to find mode of posterior.

    Args:
        seed: random seed
        log_prob_fn: log probability function
        init_params: initial parameters
        gradient_descent_iters: number of gradient descent iterations
        gradient_descent_lr: gradient descent learning rate

    Returns:
        params: sampled parameters
        losses: loss history
    """
    # find the mode of the posterior
    mode, _, losses = run_gradient_descent(
        lambda x: -log_prob_fn(x),
        init_params,
        num_mstep_iters=gradient_descent_iters,
        optimizer=optax.adam(gradient_descent_lr),
    )
    # calculate covariance matrix from hessian at mode
    mode, unravel_fn = jax.flatten_util.ravel_pytree(mode)
    ll_fn = lambda x: log_prob_fn(unravel_fn(x))
    hessian_at_mode = jax.hessian(ll_fn)(mode)
    covariance_matrix = psd_inv(-hessian_at_mode, diagonal_boost=1e-2)

    # sample from laplace approximation
    x = jr.multivariate_normal(seed, mean=mode, cov=covariance_matrix)
    return unravel_fn(x), losses


def symmetrize(A: Float[Array, "n n"]) -> Float[Array, "n n"]:
    """Symmetrize a matrix by averaging it with its transpose."""
    return (A + A.swapaxes(-1, -2)) / 2


def psd_solve(
    A: Float[Array, "n n"], 
    B: Float[Array, "n m"], 
    diagonal_boost: float = 1e-6
) -> Float[Array, "n m"]:
    """Solve the linear system Ax = B, where A is a positive semi-definite matrix.
    Args:
        A: positive semi-definite matrix
        B: right-hand side matrix
        diagonal_boost: boost to diagonal to ensure positive definiteness
    Returns:
        x: solution to the linear system
    """
    A = symmetrize(A) + diagonal_boost * jnp.eye(A.shape[-1])
    L, lower = cho_factor(A, lower=True)
    x = cho_solve((L, lower), B)
    return x


def psd_inv(A : Float[Array, "n n"], diagonal_boost: float = 1e-6) -> Float[Array, "n n"]:
    """Compute the inverse of a positive semi-definite matrix using Cholesky decomposition.
    Args:
        A: positive semi-definite matrix
        diagonal_boost: boost to diagonal to ensure positive definiteness
    Returns:
        Ainv: inverse of the matrix
    """
    Ainv = psd_solve(A, jnp.eye(A.shape[-1]), diagonal_boost=diagonal_boost)
    return symmetrize(Ainv)


def cross_sequence_mutual_information(
    sequence1: Int[Array, "n_timesteps"],
    sequence2: Int[Array, "n_timesteps"],
    mask: Bool[Array, "n_timesteps"],
    n_categories: int,
    pseudo_count: float = 1e-8,
) -> Float:
    """Compute cross-sequence mutual information.

    Args:
        sequence1: first sequence
        sequence2: second sequence
        mask: mask for valid timesteps
        n_categories: number of categories
        pseudo_count: pseudo count to add to probabilities

    Returns:
        mi: mutual information
    """
    counts = (
        (jnp.ones((n_categories, n_categories)) * pseudo_count)
        .at[sequence1, sequence2]
        .add(mask)
    )
    probs = counts / counts.sum()
    mi = (probs * jnp.log(probs / (probs.sum(1)[:, na] * probs.sum(0)[na, :]))).sum()
    return mi


def lagged_mutual_information(
    sequences: Int[Array, "n_sequences n_timesteps"],
    mask: Bool[Array, "n_sequences n_timesteps"],
    lags: Int[Array, "n_lags"],
    pseudo_count: float = 1e-8,
) -> Tuple[
    Float[Array, "n_sequences n_lags"],
    Float[Array, "n_sequences n_lags"],
    Float[Array, "n_sequences n_lags"],
]:
    """Compute mutual information at a range of lags for real data and equivalent Markov chains.
    
    Args:
        sequences: sequences from which to compute mutual information
        mask: mask indicating valid timesteps
        lags: array of temporal lags
        pseudo_count: pseudo count to use when computing mutual information

    Returns:
        real_mi: mutual information for each sequence at each lag
        markov_mi: mutual information for equivalent Markov chains at each lag
        shuff_mi: mutual information across randomly paired sequences at each lag
    """
    # get dimensions and number of categories
    n_sequences, n_timesteps = sequences.shape
    n_lags = len(lags)
    n_categories = jnp.where(mask, sequences, 0).max().item() + 1

    # simulate Markov chains
    seeds = jr.split(jr.PRNGKey(0), n_sequences)
    trans_counts = jax.vmap(count_transitions, in_axes=(0, 0, None))(sequences, mask, n_categories)
    trans_probs = trans_counts / trans_counts.sum(axis=2, keepdims=True)
    init_probs = trans_counts.sum(axis=2) / trans_counts.sum(axis=(1, 2))[:,None]
    markov_seqs = jnp.zeros((n_sequences, n_timesteps), dtype=int)
    for i in tqdm.trange(n_sequences, desc="Simulating Markov chains"):
        markov_seqs = markov_seqs.at[i].set(
            simulate_markov_chain(seeds[i], trans_probs[i], n_timesteps, init_probs[i])
        )

    # compute mutual information 
    cross_mi = jax.vmap(
        jax.jit(cross_sequence_mutual_information, static_argnums=(3, 4)),
        in_axes=(0, 0, 0, None, None),
    )
    real_mi = jnp.zeros((n_sequences, n_lags))
    markov_mi = jnp.zeros((n_sequences, n_lags))
    shuff_mi = jnp.zeros((n_sequences, n_lags))

    shuff_seqs = jnp.roll(sequences, 1, axis=0)
    for i, lag in tqdm.tqdm(enumerate(lags), total=n_lags, desc="Computing MI"):
        lagged_real_seqs = jnp.roll(sequences, lag, axis=1)
        lagged_markov_seqs = jnp.roll(markov_seqs, lag, axis=1)
        lagged_mask = mask.at[:, :lag].set(0)

        real_mi = real_mi.at[:, i].set(
            cross_mi(sequences, lagged_real_seqs, lagged_mask, n_categories, pseudo_count)
        )
        markov_mi = markov_mi.at[:, i].set(
            cross_mi(markov_seqs, lagged_markov_seqs, lagged_mask, n_categories, pseudo_count)
        )
        shuff_mi = shuff_mi.at[:, i].set(
            cross_mi(shuff_seqs, lagged_real_seqs, lagged_mask, n_categories, pseudo_count)
        )
    return real_mi, markov_mi, shuff_mi


def save_hdf5(
    filepath: str,
    save_dict: Dict[str, PyTree],
    datapath: Optional[str] = None,
    overwrite_results: bool = False,
) -> None:
    """Save a dict of pytrees to an hdf5 file. The leaves of the pytrees must
    be numpy arrays, scalars, or strings.

    Args:
        filepath: Path of the hdf5 file to create.
        save_dict: Dictionary where the values are pytrees.
        datapath: Path within hdf5 file to save the data. If None, data are saved at the root.
    """
    with h5py.File(filepath, "a") as f:
        if datapath is not None:
            _savetree_hdf5(jax.device_get(save_dict), f, datapath)
        else:
            for k, tree in save_dict.items():
                _savetree_hdf5(jax.device_get(tree), f, k)


def load_hdf5(
    filepath: str,
    datapath: Optional[str] = None,
) -> Dict[str, PyTree]:
    """Load a dict of pytrees from an hdf5 file.

    Args:
        filepath: Path of the hdf5 file to load.
        datapath: Path within hdf5 file to load data from. If None, loads from the root.

    Returns:
        save_dict: Dictionary where the values are pytrees.
    """
    with h5py.File(filepath, "r") as f:
        if datapath is None:
            return {k: _loadtree_hdf5(f[k]) for k in f}
        else:
            return _loadtree_hdf5(f[datapath])


def _savetree_hdf5(tree: PyTree, group: h5py.Group, name: str) -> None:
    """Recursively save a pytree to an h5 file group."""
    if name in group:
        del group[name]
    if isinstance(tree, np.ndarray):
        if tree.dtype.kind == "U":
            dt = h5py.special_dtype(vlen=str)
            group.create_dataset(name, data=tree.astype(object), dtype=dt)
        else:
            group.create_dataset(name, data=tree)
    elif isinstance(tree, (float, int, str)):
        group.create_dataset(name, data=tree)
    else:
        subgroup = group.create_group(name)
        subgroup.attrs["type"] = type(tree).__name__

        if isinstance(tree, (tuple, list)):
            for k, subtree in enumerate(tree):
                _savetree_hdf5(subtree, subgroup, f"arr{k}")
        elif isinstance(tree, dict):
            for k, subtree in tree.items():
                _savetree_hdf5(subtree, subgroup, k)
        else:
            raise ValueError(f"Unrecognized type {type(tree)}")


def _loadtree_hdf5(leaf: Union[h5py.Dataset, h5py.Group]) -> PyTree:
    """Recursively load a pytree from an h5 file group."""
    if isinstance(leaf, h5py.Dataset):
        data = np.array(leaf[()])
        if h5py.check_dtype(vlen=data.dtype) == str:
            data = np.array([item.decode("utf-8") for item in data])
        elif data.dtype.kind == "S":
            data = data.item().decode("utf-8")
        elif data.shape == ():
            data = data.item()
        return data
    else:
        leaf_type = leaf.attrs["type"]
        values = map(_loadtree_hdf5, leaf.values())
        if leaf_type == "dict":
            return dict(zip(leaf.keys(), values))
        elif leaf_type == "list":
            return list(values)
        elif leaf_type == "tuple":
            return tuple(values)
        else:
            raise ValueError(f"Unrecognized type {leaf_type}")


def unbatch(
    data: Array,
    keys: Union[list[str], Array],
    bounds: Int[Array, "n_segs 2"]
) -> Dict[str, Array]:
    """Invert :py:func:`state_moseq.util.batch`

    Args:
        data: Stack of segmented time-series, shape (n_segs, seg_length, ...).
        keys: Name of the time-series that each segment came from
        bounds: Start and end indices for each segment.

    Returns:
        data_dict: Dictionary mapping names to reconstructed time-series.
    """
    data_dict = {}
    for key in set(list(keys)):
        length = bounds[keys == key, 1].max()
        seq = np.zeros((int(length), *data.shape[2:]), dtype=data.dtype)
        for (s, e), d in zip(bounds[keys == key], data[keys == key]):
            seq[s:e] = d[: e - s]
        data_dict[key] = seq
    return data_dict


def batch(
    data_dict: Dict[str, Array],
    keys: Optional[list[str]] = None,
    seg_length: Optional[int] = None,
    seg_overlap: int = 30,
) -> Tuple[Array, Int[Array, "N seg_length"], Tuple[list[str], Int[Array, "N 2"]]]:
    """Stack time-series data of different lengths into a single array for batch
    processing, optionally breaking up the data into fixed length segments. The
    data is padded so that the stacked array isn't ragged. The padding
    repeats the last frame of each time-series until the end of the segment.

    Args:
        data_dict: Dictionary of time-series, each of shape (T, ...).
        keys: Optional list of keys to control order and inclusion of time-series.
        seg_length: Length of each segment. Defaults to max sequence length.
        seg_overlap: Overlap between segments in frames.

    Returns:
        data: Stacked data array, shape (N, seg_length, ...).
        mask: Binary mask for valid data (1 = valid, 0 = padding), shape (N, seg_length).
        metadata: Tuple (keys, bounds), identifying sources and segment positions.
    """
    if keys is None:
        keys = sorted(data_dict.keys())
    Ns = [len(data_dict[key]) for key in keys]
    if seg_length is None:
        seg_length = np.max(Ns)

    stack, mask, keys_out, bounds = [], [], [], []
    for key, N in zip(keys, Ns):
        for start in range(0, N, seg_length):
            arr = data_dict[key]
            end = min(start + seg_length + seg_overlap, N)
            pad_length = seg_length + seg_overlap - (end - start)
            padding = np.repeat(arr[end - 1 : end], pad_length, axis=0)
            mask.append(np.hstack([np.ones(end - start), np.zeros(pad_length)]))
            stack.append(np.concatenate([arr[start:end], padding], axis=0))
            keys_out.append(key)
            bounds.append((start, end))

    stack = np.stack(stack)
    mask = np.stack(mask)
    metadata = (np.array(keys_out), np.array(bounds))
    return stack, mask, metadata


def get_durations(
    states_dict: Dict[str, Int[Array, "n_timesteps"]]
) -> Int[Array, "n_durations"]:
    """Get durations of high-level states.

    Args:
        states_dict: Dictionary of high-level state sequences.

    Returns:
        durations: Times between high-level state transitions (across all sequences).

    Examples:
        >>> states_dict = {
        ...     'name1': np.array([1, 1, 2, 2, 2, 3]),
        ...     'name2': np.array([0, 0, 0, 1]),
        ... }
        >>> get_durations(states_dict)
        array([2, 3, 1, 3, 1])
    """
    stateseq_flat = np.hstack(list(states_dict.values()))
    stateseq_padded = np.hstack([[-1], stateseq_flat, [-1]])
    changepoints = np.diff(stateseq_padded).nonzero()[0]
    return changepoints[1:] - changepoints[:-1]


def sample_instances(
    states_dict: Dict[str, Int[Array, "n_timesteps"]],
    num_instances: int,
) -> Dict[int, List[Tuple[str, int, int]]]:
    """Randomly sample instances of each state.

    Args:
        states_dict: Dictionary of state sequences.
        num_instances: Number of instances per state.

    Returns:
        sampled_instances: Dictionary mapping state index to instances.
    """
    state_ixs = np.unique(np.hstack(list(states_dict.values())))
    all_instances = {state_ix: [] for state_ix in state_ixs}

    for key, stateseq in states_dict.items():
        transitions = np.nonzero(stateseq[1:] != stateseq[:-1])[0] + 1
        starts = np.insert(transitions, 0, 0)
        ends = np.append(transitions, len(stateseq))
        for s, e, state in zip(starts, ends, stateseq[starts]):
            all_instances[state].append((key, s, e))

    sampled_instances = {}
    for state_ix, instances in all_instances.items():
        subset = np.random.permutation(len(instances))[:num_instances]
        sampled_instances[state_ix] = [instances[i] for i in subset]

    return sampled_instances


def _concatenate_stateseqs(states_dict):
    """Concatenate high-level state sequences from a dictionary into a single array."""
    return np.hstack([states_dict[key] for key in sorted(states_dict.keys())]).astype(int)


def get_frequencies(
    states_dict: Dict[str, Int[Array, "n_timesteps"]],
    num_states: Optional[int] = None,
    runlength: bool = False,
) -> Float[Array, "n_states"]:
    """Get frequencies for a batch of high-level state sequences.

    Args:
        states_dict: Dictionary of high-level state sequences.
        num_states: Total number of states. If None, inferred from data.
        runlength: If True, count only the first timepoint of each run of a state.
    
    Returns:
        frequencies: Frequency of each state across all state sequences

    Examples:
        >>> states_dict = {
            'name1': np.array([1, 1, 2, 2, 2, 3]),
            'name2': np.array([0, 0, 0, 1])}
        >>> get_frequencies(states_dict, runlength=True)
        array([0.2, 0.4, 0.2, 0.2])
        >>> get_frequencies(states_dict, runlength=False)
        array([0.3, 0.3, 0.3, 0.1])
    """
    stateseq_flat = _concatenate_stateseqs(states_dict)

    if num_states is None:
        num_states = np.max(stateseq_flat) + 1

    if runlength:
        state_onsets = np.pad(np.diff(stateseq_flat).nonzero()[0] + 1, (1, 0))
        stateseq_flat = stateseq_flat[state_onsets]

    counts = np.bincount(stateseq_flat, minlength=num_states)
    frequencies = counts / counts.sum()
    return frequencies


def get_adjusted_rand(
    states_dict1: Dict[str, Int[Array, "n_timesteps"]],
    states_dict2: Dict[str, Int[Array, "n_timesteps"]],
    downsample: int = 10,
) -> float:
    """Compute the adjusted Rand index between two sets of high-level state sequences.
    Args:
        states_dict1: First dictionary of high-level state sequences.
        states_dict2: Second dictionary of high-level state sequences.
        downsample: Downsampling factor to reduce the length of the sequences.
    Returns:
        adjusted_rand_index: Adjusted Rand index between the two sets of state sequences.
    """
    seq1 = _concatenate_stateseqs(states_dict1)[::downsample]
    seq2 = _concatenate_stateseqs(states_dict2)[::downsample]
    return  adjusted_rand_score(seq1, seq2)


import jax
import jax.numpy as jnp
import jax.random as jr
from jax.scipy.stats import norm, dirichlet
from jaxtyping import Array, Float, Int
from jax.nn import softmax
import tqdm
from typing import Tuple
from functools import partial
from dynamax.hidden_markov_model import (
    hmm_filter,
    hmm_smoother,
    hmm_posterior_mode,
    hmm_posterior_sample,
    parallel_hmm_filter,
    parallel_hmm_smoother,
    parallel_hmm_posterior_sample,
)

from .util import (
    simulate_markov_chain,
    lower_dim,
    raise_dim,
    sample_laplace,
    count_transitions
)

na = jnp.newaxis


def estimate_emission_params(
    sufficient_stats: Float[Array, "n_states n_syllables n_syllables"],
) -> Tuple[
    Float[Array, "n_syllables n_syllables-1"], Float[Array, "n_states-1 n_syllables-1"]
]:
    """Estimate emission parameters from transition counts."""
    logits = jnp.log(sufficient_stats + 1e-2)
    emission_base_est = lower_dim(logits.mean(0), 1)
    emission_biases_est = (logits - logits.mean(0)[na]).mean(1)
    emission_biases_est = lower_dim(lower_dim(emission_biases_est, 1), 0)
    return emission_base_est, emission_biases_est


def get_syllable_trans_probs(
    emission_base: Float[Array, "n_syllables n_syllables-1"],
    emission_biases: Float[Array, "n_states-1 n_syllables-1"],
) -> Float[Array, "n_states n_syllables n_syllables"]:
    """Compute transition probabilities between syllables."""
    emission_base = raise_dim(emission_base, 1)
    emission_biases = raise_dim(raise_dim(emission_biases, 0), 1)
    logits = emission_base[na] + emission_biases[:, na]
    return softmax(logits, axis=-1)


def obs_log_likelihoods(
    data: dict,
    params: dict,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Compute log likelihoods of observations for each hidden state."""

    n_sequences = data["syllables"].shape[0]
    n_states = params["trans_probs"].shape[0]

    log_syllable_trans_probs = jnp.log(
        get_syllable_trans_probs(
            params["emission_base"],
            params["emission_biases"],
        )
    )
    log_likelihoods = jax.vmap(
        lambda T: T[data["syllables"][:, :-1], data["syllables"][:, 1:]]
    )(log_syllable_trans_probs)

    log_likelihoods = jnp.concatenate(
        [jnp.zeros((n_states, n_sequences, 1)), log_likelihoods], axis=2
    )
    return log_likelihoods.transpose((1, 2, 0)) * data["mask"][:, :, na]


def log_params_prob(
    params: dict,
    hypparams: dict,
) -> Float:
    """Compute the log probability of the parameters based on their priors."""

    n_states = params["trans_probs"].shape[0]

    # prior on emission base parameters
    emission_base_log_prob = norm.logpdf(
        params["emission_base"] / hypparams["emission_base_sigma"]
    ).sum()

    # prior on emission bias parameters
    emission_biases_log_prob = norm.logpdf(
        params["emission_biases"] / hypparams["emission_biases_sigma"]
    ).sum()

    # prior on transition parameters
    trans_probs_log_prob = jax.vmap(dirichlet.logpdf)(
        params["trans_probs"],
        jnp.eye(n_states) * hypparams["trans_kappa"] + hypparams["trans_beta"],
    ).sum()

    return emission_base_log_prob + emission_biases_log_prob + trans_probs_log_prob


def log_joint_prob(
    data: dict,
    params: dict,
    hypparams: dict,
) -> Float:
    """Compute the log joint probability of the data and parameters."""
    return marginal_loglik(data, params) + log_params_prob(params, hypparams)


@partial(jax.jit, static_argnums=(3,))
def resample_states(
    seed: Float[Array, "2"],
    data: dict,
    params: dict,
    parallel: bool = False,
) -> Tuple[Int[Array, "n_sequences n_timesteps"], Float]:
    """Resample hidden states from their posterior distribution.

    Args:
        seed: random seed
        data: data dictionary
        params: parameters dictionary
        parallel: whether to use parallel message passing

    Returns:
        states: resampled hidden states
        marginal_loglik: marginal log likelihood of the data
    """
    n_states = params["trans_probs"].shape[0]
    seeds = jr.split(seed, data["syllables"].shape[0])

    sample_fn = parallel_hmm_posterior_sample if parallel else hmm_posterior_sample
    marginal_logliks, states = jax.vmap(sample_fn, in_axes=(0, None, None, 0))(
        seeds,
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    )
    return states, marginal_logliks.sum()


def fit_gibbs(
    data: dict,
    hypparams: dict,
    init_params: dict,
    init_states: Int[Array, "n_sequences n_timesteps"] = None,
    seed: Float[Array, "2"] = jr.PRNGKey(0),
    num_iters: Int = 100,
    parallel: bool = False,
) -> Tuple[dict, Float[Array, "num_iters"]]:
    """Fit a model using Gibbs sampling.

    Args:
        data: data dictionary
        hypparams: hyperparameters dictionary
        init_params: initial parameters directionary
        init_states: initial hidden states (optional)
        seed: random seed
        num_iters: number of iterations
        parallel: whether to use parallel message passing

    Returns:
        params: fitted parameters dictionary
        log_joints: log joint probability of the data and parameters recorded at each iteration
    """
    if init_states is None:
        states, _ = resample_states(seed, data, init_params, parallel)

    log_joints = []
    params = init_params
    for _ in tqdm.trange(num_iters):
        seed, subseed = jr.split(seed)
        params, gd_losses = resample_params(subseed, data, params, states, hypparams)
        states, marginal_loglik = resample_states(seed, data, params, parallel)
        log_joints.append(marginal_loglik + log_params_prob(params, hypparams))
    return params, states, jnp.array(log_joints)


def initialize_params(
    data: dict,
    hypparams: dict,
    states: Int[Array, "n_sequences n_timesteps"] = None,
    seed: Float[Array, "2"] = jr.PRNGKey(0),
) -> dict:
    """Initialize parameters by sampling from their prior distribution or using
    provided states.

    Args:
        data: data dictionary
        hypparams: hyperparameters dictionary
        states: states used for initializing the parameters (optional)
        seed: random seed
    """
    if states is not None:
        params = resample_params(seed, data, states, hypparams)
    else:
        params = random_params(seed, hypparams)
    return params


def fit_gradient_descent(
    data: dict,
    hypparams: dict,
    init_params: dict,
    num_iters: Int = 100,
    learning_rate: Float = 1e-3,
) -> Tuple[dict, Float[Array, "num_iters"]]:
    """Fit a model using gradient descent.

    Args:
        data: data dictionary
        hypparams: hyperparameters dictionary
        init_params: initial parameters directionary
        num_iters: number of iterations
        learning_rate: learning rate for gradient descent

    Returns:
        params: fitted parameters dictionary
        log_joints: log joint probability of the data and parameters recorded at each iteration
    """
    loss_fn = lambda params: -log_joint_prob(data, params, hypparams)
    params, losses = gradient_descent(loss_fn, init_params, learning_rate, num_iters)
    log_joints = -losses
    return params, log_joints


def marginal_loglik(
    data: dict,
    params: dict,
    parallel: bool = False,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Estimate marginal log likelihood of the data"""
    filter_fn = parallel_hmm_filter if parallel else hmm_filter
    n_states = params["trans_probs"].shape[0]
    mll = jax.vmap(filter_fn, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    ).marginal_loglik.sum()
    return mll


def smoothed_states(
    data: dict,
    params: dict,
    parallel: bool = False,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Estimate marginals of hidden states using forward-backward algorithm."""
    smoother_fn = parallel_hmm_smoother if parallel else hmm_smoother
    n_states = params["trans_probs"].shape[0]
    return jax.vmap(smoother_fn, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    ).smoothed_probs


def filtered_states(
    data: dict,
    params: dict,
    parallel: bool = False,
) -> Float[Array, "n_sequences n_timesteps n_states"]:
    """Estimate marginals of hidden states using forward-backward algorithm."""
    filter_fn = parallel_hmm_filter if parallel else hmm_filter
    n_states = params["trans_probs"].shape[0]
    return jax.vmap(filter_fn, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    ).filtered_probs


def predicted_states(
    data: dict,
    params: dict,
) -> Float[Array, "n_sequences n_timesteps"]:
    """Predict hidden states using Viterbi algorithm."""
    n_states = params["trans_probs"].shape[0]
    return jax.vmap(hmm_posterior_mode, in_axes=(None, None, 0))(
        jnp.ones(n_states) / n_states,
        params["trans_probs"],
        obs_log_likelihoods(data, params),
    )


def random_params(
    seed: Float[Array, "2"],
    hypparams: dict,
) -> dict:
    """Generate random model parameters.

    emission_base ~ Normal(0, emission_base_sigma)
    emission_biases ~ Normal(0, emission_biases_sigma)
    trans_probs ~ Dirichlet(trans_beta + trans_kappa * I)

    Args:
        seed: random seed
        hypparams: hyperparameters dictionary

    Returns:
        params: parameters dictionary
    """
    n_syllables = hypparams["n_syllables"]
    n_states = hypparams["n_states"]
    seeds = jr.split(seed, 3)

    emission_base = (
        jr.normal(seeds[0], shape=(n_syllables, n_syllables - 1))
        * hypparams["emission_base_sigma"]
    )

    emission_biases = (
        jr.normal(seeds[1], shape=(n_states - 1, n_syllables - 1))
        * hypparams["emission_biases_sigma"]
    )

    trans_probs = jax.vmap(jr.dirichlet)(
        jr.split(seeds[2], n_states),
        jnp.eye(n_states) * hypparams["trans_kappa"] + hypparams["trans_beta"],
    )

    return {
        "emission_base": emission_base,
        "emission_biases": emission_biases,
        "trans_probs": trans_probs,
    }


def simulate(
    seed: Float[Array, "2"],
    params: dict,
    n_timesteps: Int,
    n_sequences: Int,
) -> Tuple[
    Int[Array, "n_sequences n_timesteps"], Int[Array, "n_sequences n_timesteps"]
]:
    """Simulate data from the model.

    Args:
        seed: random seed
        params: parameters dictionary
        n_timesteps: number of timesteps to simulate
        n_sequences: number of sessions to simulate

    Returns:
        states: simulated states
        syllables: simulated syllables
    """
    seeds = jr.split(seed, 3)

    states = jax.vmap(simulate_markov_chain, in_axes=(0, None, None))(
        jr.split(seeds[0], n_sequences),
        params["trans_probs"],
        n_timesteps,
    )
    syllable_trans_probs = get_syllable_trans_probs(
        params["emission_base"],
        params["emission_biases"],
    )[states]

    syllables = jax.vmap(simulate_markov_chain, in_axes=(0, 0, None))(
        jr.split(seeds[1], n_sequences), syllable_trans_probs, n_timesteps
    )
    return states, syllables


def resample_params(
    seed: Float[Array, "2"],
    data: dict,
    states: Int[Array, "n_sequences n_timesteps"],
    hypparams: dict,
    params: dict = None,
) -> Tuple[dict, Float[Array, "gradient_descent_iters"]]:
    """Resample parameters from their posterior distribution. Emission parameters are
    resampled using a Laplace approximation; the mode is found using gradient descent.

    Args:
        seed: random seed
        data: data dictionary
        states: hidden states
        hypparams: hyperparameters dictionary
        params: parameters dictionary (optional; used for initializing gradient descent)

    Returns:
        params: parameters dictionary
        losses: losses recorded during gradient descent
    """
    seeds = jr.split(seed, 2)

    if params is None:
        init_emission_base = None
        init_emission_biases = None
    else:
        init_emission_base = params["emission_base"]
        init_emission_biases = params["emission_biases"]

    emission_params, gd_losses = resample_emission_params(
        seeds[1],
        data["syllables"],
        data["mask"],
        states,
        hypparams["n_states"],
        hypparams["n_syllables"],
        hypparams["emission_base_sigma"],
        hypparams["emission_biases_sigma"],
        hypparams["emission_gd_iters"],
        hypparams["emission_gd_lr"],
        init_emission_base,
        init_emission_biases,
    )
    trans_probs = resample_trans_probs(
        seeds[2],
        data["mask"],
        states,
        hypparams["n_states"],
        hypparams["trans_beta"],
        hypparams["trans_kappa"],
    )
    params = {
        "emission_base": emission_params[0],
        "emission_biases": emission_params[1],
        "trans_probs": trans_probs,
    }
    return params, gd_losses


@partial(jax.jit, static_argnums=(4,5,8,))
def resample_emission_params(
    seed: Float[Array, "2"],
    syllables: Int[Array, "n_sequences n_timesteps"],
    mask: Int[Array, "n_sequences n_timesteps"],
    states: Int[Array, "n_sequences n_timesteps"],
    n_states: int,
    n_syllables: int,
    emission_base_sigma: Float,
    emission_biases_sigma: Float,
    gradient_descent_iters: Int = 100,
    gradient_descent_lr: Float = 1e-3,
    init_emission_base: Float[Array, "n_syllables n_syllables-1"] = None,
    init_emission_biases: Float[Array, "n_states-1 n_syllables-1"] = None,
) -> Tuple[
    Tuple[
        Float[Array, "n_syllables n_syllables-1"],
        Float[Array, "n_states-1 n_syllables-1"],
    ],
    Float[Array, "gradient_descent_iters"],
]:
    """Resample emission parameters from their posterior distribution.

    Args:
        seed: random seed
        syllables: syllable observations
        mask: mask of valid observations
        states: hidden states
        n_states: number of hidden states
        n_syllables: number of syllables
        emission_base_sigma: emission base standard deviation
        emission_biases_sigma: emission biases standard deviation
        gradient_descent_iters: number of gradient descent iterations
        gradient_descent_lr: gradient descent loss rate
        init_emission_base: initial emission base parameters (optional)
        init_emission_biases: initial emission biases parameters (optional)

    Returns:
        emission_base: posterior emission base parameters
        emission_biases: posterior emission biases parameters
        losses: losses recorded during gradient descent
    """
    sufficient_stats = (
        jnp.zeros((n_states, n_syllables, n_syllables))
        .at[states[:, 1:], syllables[:, :-1], syllables[:, 1:]]
        .add(mask[:, 1:])
    )

    def log_prob_fn(args):
        emission_base, emission_biases = args
        syllable_trans_probs = get_syllable_trans_probs(emission_base, emission_biases)

        emission_base = raise_dim(emission_base, 1)
        emission_biases = raise_dim(raise_dim(emission_biases, 0), 1)
        prior_log_prob = (
            norm.logpdf(emission_base / emission_base_sigma).sum()
            + norm.logpdf(emission_biases / emission_biases_sigma).sum()
        )
        syllables_log_prob = (jnp.log(syllable_trans_probs) * sufficient_stats).sum()
        return prior_log_prob + syllables_log_prob

    if init_emission_base is None or init_emission_biases is None:
        init_emission_params = estimate_emission_params(sufficient_stats)
    else:
        init_emission_params = (init_emission_base, init_emission_biases)
    (emission_base, emission_biases), losses = sample_laplace(
        seed,
        log_prob_fn,
        init_emission_params,
        gradient_descent_iters,
        gradient_descent_lr,
    )
    return (emission_base, emission_biases), losses


@partial(jax.jit, static_argnums=(3,))
def resample_trans_probs(
    seed: Float[Array, "2"],
    mask: Int[Array, "n_sequences n_timesteps"],
    states: Int[Array, "n_sequences n_timesteps"],
    n_states: int,
    beta: Float,
    kappa: Float,
) -> Float[Array, "n_states n_states"]:
    """Resample transition probabilities from their posterior distribution.

    Args:
        seed: random seed
        mask: mask of valid observations
        states: hidden states
        n_states: number of hidden states
        beta: Dirichlet concentration parameter
        kappa: Dirichlet concentration parameter

    Returns:
        trans_probs: posterior transition probabilities
    """
    trans_counts = jax.vmap(count_transitions, in_axes=(0, 0, None))(states, mask, n_states).sum(0)
    trans_probs = jax.vmap(jr.dirichlet)(
        jr.split(seed, n_states), trans_counts + beta + jnp.eye(n_states) * kappa
    )
    return trans_probs
